@echo off
rem To be used on MS-Windows for Msys2 with the 64 bit MinGW compiler.
rem Adjust the "c:\msys64" part to match your installation.
@echo on

set PATH=c:\msys64\mingw64\bin;c:\msys64\usr\bin;%PATH%
