" Vim filetype plugin file
" Language:	FreeBasic
" Maintainer:	Doug Kearns <dougkearns@gmail.com>
" Last Change:	2015 Jan 10

if exists("b:did_ftplugin")
  finish
endif
let b:did_ftplugin = 1

runtime! ftplugin/basic.vim

" vim: ts=8
